; 
    include \masm32\include\masm32rt.inc
; 
.586p

    .data
    ;;;;;;; CONFIGURATION ;;;;;;;
    ;; gu_* stands for global user preference. These are meant to be adjusted
    ;; independently of any other code.
      ; What priority should the test driver run at?
      gu_priorityClass textequ <HIGH_PRIORITY_CLASS>

      ; Number of successive runs through the test code that must take the
      ; exact same number of CPU ticks. This setting allows us to drop
      ; spurious results like context switches, interrupts, CPU throttling,
      ; cache misses and whatever other "one-off" events occur.
      ; NOTE: this number will be automatically adjusted by the test driver
      ; on an as needed basis to ensure test results are stable and timely.
      ; This is essentially the starting point for the driver.
      gu_default_sample_batch_size equ 8

      ; The driver may raise the sample_batch_size no higher than this number.
      ; Some combinations of opcodes are super stable and give spurious results
      ; at lower settings, so the driver will attempt to adjust the testing to
      ; the code sequence under test. (Currently UNIMPLEMENTED)
      gu_max_sample_batch_size equ 128

      ; How many times do we attempt to get a stable sample reading before
      ; giving up and adjusting downwards the sample_batch_size?
      gu_sample_retry_attempts equ 100000

      ; Total number of samples to print out. In "production" versions
      ; this ought to be some small number like 8 or 16.
      gu_total_samples equ 500


      ; Idicates how many times the test code inside of gu_testfunction should
      ; loop. This helps to amortize the cost of calling the function as well
      ; as allows the code under test, especially if short, to make full use
      ; of cache.
      ;
      ; It is very important that test functions include this constant as
      ; the loop variable or that you set this to 1!
      gu_short_inner_loop_count equ 4
      gu_inner_loop_count equ 32
      gu_long_inner_loop_count equ 128
    ;;;;;;; END CONFIG ;;;;;;;;;;
    

      align 16
      g_run_cache dd gu_max_sample_batch_size dup(0)
      align 16
      g_steps dd gu_default_sample_batch_size-1
      align 16
      g_result_timings dd gu_total_samples dup(0)

      align 16
      gu_testfunction dd ?

      g_testfunction_call_count dd 0
      g_isbaseline db 0 ; is the current test running a baseline or not?

      align 16
    
     ;; record information about functions under test
      g_baseline_min dd ?
      g_baseline_max dd ?
      g_baseline_mean dd ?
      g_baseline_range dd ?
      g_baseline_variance dd ?
      g_baseline_fpMean real8 ?
      g_baseline_inner_loop_count dd ?

    
      g_test_min dd ?
      g_test_max dd ?
      g_test_mean dd ?
      g_test_range dd 1000
      g_test_variance dd ?
      g_test_fpMean real8 ?
      g_test_sdev real8 ?
      g_test_mean_r8 real8 ?
    .code

start:
   ; 
    invoke GetCurrentProcess
    invoke SetPriorityClass, eax, gu_priorityClass
    invoke GetCurrentThread
    invoke SetThreadPriority, eax, THREAD_PRIORITY_TIME_CRITICAL
    invoke GetCurrentProcess
    invoke SetProcessAffinityMask,eax,1
    invoke Sleep, 50

    ;; setting up the fpu
    finit
    fstcw WORD PTR [esp+4]
    mov   ax,      [esp+4]
    and   ax,      0f3ffh
    or    ax,      0c00h
    mov   [esp+4], ax
    fldcw [esp+4]
    

    cls
    print "Spinning up the processor.",13,10
    mov ecx, 31
    l_spinup:
      call init_spinup
      dec ecx
    jns l_spinup

    call runner

    inkey
    exit

;; eax has pointer to statistics
;; [eax+0] = min, [eax+4] = max, [eax+8] = mean, [eax+12] = range, [eax+16] = variance.
print_stat_listing proc
    push ebx
    mov ebx, eax
    
    lea eax, g_test_fpMean
    call print_float
    print " +/- "
    lea eax, g_test_sdev
    call print_float
    print " ticks"
    print chr$(13,10)
    pop ebx
print_stat_listing endp
; 
;; Run the test driver. This will call main and after that returns
;; uses the contents of g_result_timings to compute various statistics.
run_driver proc
    invoke Sleep,10
    push ebx
    push esi
    push edi

    mov [g_testfunction_call_count], 0 ; initialize to zero
    
    call main
    
    ;print " ",13,10
    ;print "Min:      "
    call calc_minimum
    push eax
    test [g_isbaseline],1
    jne baseline_min
    mov [g_test_min], eax
    jmp test_min
  baseline_min:
    mov [g_baseline_min], eax
  test_min:
    ;print ustr$(eax)

    ;print " ticks, ";,13,10
    ;print "Max:      "
    call calc_maximum
    push eax
    test [g_isbaseline],1
    jne baseline_max
    mov [g_test_max], eax
    jmp test_max
  baseline_max:
    mov [g_baseline_max], eax
  test_max:
    ;print ustr$(eax)

    ;print " ticks, ";,13,10
    ;print "Range:    "
    pop eax ; max
    pop ecx ; min
    sub eax, ecx
    test [g_isbaseline],1
    jne baseline_range
    mov [g_test_range], eax
    jmp test_range
  baseline_range:
    mov [g_baseline_range], eax
  test_range:
    ;print ustr$(eax)

call calc_fpMean
    test [g_isbaseline],1
    jne baseline_fpMean
    fstp REAL8 PTR [g_test_fpMean]
    jmp test_fpMean
  baseline_fpMean:
    fstp REAL8 PTR [g_baseline_fpMean]
  test_fpMean:

    ;print " ticks ",13,10
    ;print "Mean:     "
    call calc_mean
    push eax
    test [g_isbaseline],1
    jne baseline_mean
    mov [g_test_mean], eax
    jmp test_mean
  baseline_mean:
    mov [g_baseline_mean], eax
  test_mean:
    ;print ustr$(eax)

    ;print " ticks, ";,13,10
    ;print "Variance: "
    pop eax
    call calc_variance
    test [g_isbaseline],1
    jne baseline_variance
    mov [g_test_variance], eax
    jmp test_variance
  baseline_variance:
    mov [g_baseline_variance], eax
  test_variance:
    ;print ustr$(eax)

    ;print " ticks^2 ",13,10

    ;print "Batch Size: "
    mov eax, [g_steps]
    add eax, 1
    ;print ustr$(eax)
    ;print "      , ";,13,10
    

    ;print "Call Count: "
    mov eax, [g_testfunction_call_count]
    ;print ustr$(eax)
    ;print " calls",13,10
    pop edi
    pop esi
    pop ebx
    ret
run_driver endp



;; Our job here is to loop through the test function gu_total_samples times.
;; Timings are placed into the g_result_timings global buffer.
main proc
    ; now run for real, printing each result to standard out.
    mov ecx, gu_total_samples-1
    align 16
l:
    call do_sample_run
    mov [g_result_timings+ecx*4], eax
    dec ecx
    jns l

    ;call print_datapoints ;; UNCOMMENT TO PRINT DATAPOINTS
    ret
main endp



print_datapoints proc
    print chr$(13,10)
    print "{"
    mov ecx, gu_total_samples-1
  l_final_print:
    mov eax, [g_result_timings+ecx*4]
    push eax
    call print_number
    dec ecx
    jns l_final_print

    print "}",13,10
    ret
print_datapoints endp



;; take samples until the cpu takes the same amount of time to do N runs.
;; RETURN: EAX=ticks taken.
align 16
do_sample_run proc
    push ecx
    push ebp

reset_and_retry:
    mov ebp, [g_steps]
    mov ecx, gu_sample_retry_attempts
    ;invoke Sleep,1000
    align 16
retry:
    
    dec ecx
    jz stepdown
    push ecx
    call spinup

    mov edi, ebp
    align 16
l:
    call timeit
    mov [g_run_cache+edi*4], eax
    dec edi
    jns l

    mov edi, [g_testfunction_call_count]
    add edi, ebp
    mov [g_testfunction_call_count], edi

    mov edi, ebp
    mov eax, [g_run_cache+edi*4]
    pop ecx
l_check_equality:
    cmp [g_run_cache+edi*4], eax
    jne retry
    dec edi
    jns l_check_equality

    pop ebp
    pop ecx
    ret
 stepdown:                ; We can't get any stable numbers passing muster.
    sub [g_steps], 1      ; So step down the number of required readings.
    jmp reset_and_retry
do_sample_run endp



;; Does not visibly change any registers.
;; Simply spins up the CPU in preperation of "real" timing tests.
align 16
spinup proc
    mov eax, 10
    push ecx
    mov ecx, 10000
l:
    ;imul eax,eax,10
    dec ecx
    jns l
    pop ecx
    ret
spinup endp


align 16
init_spinup proc
    mov eax, 10
    push ecx
    mov ecx, 777777777
l:
    dec ecx
    jns l
    pop ecx
    ret
init_spinup endp


align 16
timeit proc
    xor eax, eax               ; start by warming up the CPU
    cpuid                      ; We call CPUID 6 times
    xor eax, eax               ; to ensure that the amount of time
    cpuid                      ; required to execute this is constant.
    xor eax, eax               ; Intel suggests doing this 3 times,
    cpuid                      ; but experimentally, 6 does a better job.
    xor eax, eax
    cpuid
    xor eax, eax
    cpuid
    xor eax, eax
    cpuid
    ;rdtsc 
    ; now go for real!
    xor eax, eax
    cpuid
    rdtsc
   
    push eax
    push edx 
    call [gu_testfunction]

    xor eax, eax
    cpuid
    rdtsc
    push eax
    push edx
    call calc_rdtsc_delta

    ret

timeit endp

;;; integer to take the integer squareroot of is in [esp]
calc_isqrt proc
    pop ecx
    fild DWORD PTR [esp]
    fsqrt ; st0
    fistp DWORD PTR [esp]
    pop eax
    push ecx
    ret
calc_isqrt endp

;; result returned in fp(0).
calc_relative_deviation proc
    pop ecx
    ;pop eax ; mean
    ;pop edx ; standard_deviation
    ;finit
    fild DWORD PTR [esp] ; mean (sp1) 
    fild DWORD PTR [esp-4] ; variance
    fsqrt ; variance -> standard deviation
    fdiv st,st(1)  ; standard deviation/mean
    fstp st(1)        ; move result up one to clear floating point stack.
    pop eax
    pop eax
    push ecx ; reset normal return pointer, minus the two arguments we put on stack.
    ret
calc_relative_deviation endp

calc_rdtsc_delta proc
    pop esi
    pop edx
    pop eax
    pop ecx

    cmp ecx, edx ; if RDTSC wrapped around, the high 32 bits will not be equal.
    jne wraparound
    
    pop ecx
    sub eax, ecx
    push esi
    ret
    
    align 16
wraparound:
    mov eax, 0
    pop ecx
    push esi
    ret
calc_rdtsc_delta endp



;; destroys the eax, edx register, not used for return result.
align 16
print_number  proc
    pop ebp
    pop eax
    push ecx
    print ustr$(eax)
    print ", "
    pop ecx
    push ebp
    ret
print_number endp

;;;; Compute the minimum of result timings
;; result in EAX
calc_minimum proc
    mov ecx, gu_total_samples-1
    mov eax, [g_result_timings+ecx*4]
  l:
    mov edx, [g_result_timings+ecx*4]
    cmp eax, edx
    jl skip
    mov eax, edx
  skip:
    dec ecx
    jns l
    ret
calc_minimum endp

;;;; Compute the maximum of result timings
;; result in EAX
calc_maximum proc
    mov ecx, gu_total_samples-1
    mov eax, [g_result_timings+ecx*4]
  l:
    mov edx, [g_result_timings+ecx*4]
    cmp eax, edx
    jg skip
    mov eax, edx
  skip:
    dec ecx
    jns l
    ret
calc_maximum endp

;;;; Compute the mean of result timings
;; result in EAX
calc_mean proc
    mov ecx, gu_total_samples-1
    xor eax, eax
  l:
    mov edx, [g_result_timings+ecx*4]
    add eax, edx
    dec ecx
    jns l
    mov ecx, gu_total_samples
    xor edx, edx
    div ecx
    ret
calc_mean endp

;; result returned in FT0.
;; pass 
calc_fpMean proc
    xor eax,eax
    mov ecx, gu_total_samples-1
  l:
    mov edx, [g_result_timings+ecx*4]
    add eax, edx
    dec ecx
    jns l
    mov [esp+4],eax  ; store sum of all samples.
    fild DWORD PTR [esp+4]  ; load the total as a floating point
    mov ecx, [gu_total_samples]
    mov [esp+4], ecx                  ; now place count of samples*
    fidiv DWORD PTR [esp+4]  ; result now in st0
    ret
calc_fpMean endp

;;; result is eax, remainder in edx
calc_variance proc
    push edi ; callie save
    push esi ; registers
    push ebx ; all need pushed
    mov edi, eax ; store `mean` in edi for use computing variance.
    xor edx, edx
    xor eax, eax
    mov esi, gu_total_samples-1 ; tell it where we are starting...

    align 16
l:
    mov ecx, [esi*4+g_result_timings]
    sub ecx, edi      ; value - mean of all samples
    imul ecx, ecx     ; square the difference
    mov ebx, ecx
    sar ebx, 31       ; shift arithmatic right by 31
    add eax, ecx
    adc edx, ebx
    sub esi, 1
    jns l

    ;; At this point eax=low 32bits and edx=high 32 bits
    ;; esi, ebx, ecx, ebp, edi are all free for use.
    mov ecx, gu_total_samples ; number of samples to divide by
    ;;;; TODO, fix this to do proper 64 bit division!
    xor edx, edx ; we are only doing 32 bit division... discarding all high bits!
    idiv ecx

    ; now result is in eax, remainder is in edx (we usually don't care about this)
    
    pop ebx  ; restore
    pop esi  ; callie save
    pop edi  ; registers
    ret
calc_variance endp


;; assume pointer to float is in eax
;; float is REAL8
print_float PROC
    push ebx
    push ecx
    push edx
    fld REAL8 PTR [eax]            ; Load, store as integer, and print
    fist DWORD PTR [esp+4]        ; the integer portion of the number.
    mov eax, [esp+4]
    print ustr$(eax)

    print "."                      ; Print decimal point marker.

    mov ebx, 3
  l:
    ;;; now prepare to print first 3 digits after the decimal point.
    mov ecx, 10
    mov [esp+4], ecx
    
    fimul DWORD PTR [esp+4]
    fist DWORD PTR [esp+4]
    mov eax, [esp+4]
    mov ecx, 10
    xor edx, edx
    div ecx
    ;; we want remainder, which is in edx.
    print ustr$(edx)

    dec ebx
    jns l
    fistp DWORD PTR [esp+4]
    pop edx
    pop ecx
    pop ebx
    ret
print_float ENDP

RunTest PROTO   :LPVOID,:LPSTR
RunTestImpl PROTO   :LPVOID,:LPSTR
RunBaseline PROTO   :LPVOID,:LPSTR,:DWORD

RunTest PROC lpfnFunction:LPVOID,lpszString:LPSTR
    print lpszString
    print ": "
  ;retry:
    mov [g_isbaseline], 0
    invoke RunTestImpl,lpfnFunction,lpszString
  ;  cmp [g_test_range], 100
  ;  jg retry
    mov eax, [g_baseline_mean]
    sub [g_test_min], eax
    sub [g_test_mean], eax
    mov edx, [g_baseline_variance]
    sub [g_test_max], eax
    add [g_test_variance], edx

    lea edx, g_test_fpMean
    fld REAL8 PTR [edx]
    ;lea edx, g_baseline_fpMean
    fsub REAL8 PTR [g_baseline_fpMean]
    fidiv [g_baseline_inner_loop_count]
    fstp REAL8 PTR [g_test_fpMean]
    

    fild DWORD PTR [g_test_variance]
    fsqrt
    fstp REAL8 PTR [g_test_sdev]

    fild DWORD PTR [g_test_mean]
    fidiv [g_baseline_inner_loop_count]
    fstp REAL8 PTR [g_test_mean_r8]
    fld REAL8 PTR [g_test_sdev]
    fidiv [g_baseline_inner_loop_count]
    fstp REAL8 PTR [g_test_sdev]

    ;printf("SDev=(%f) mean=(%f)",g_test_sdev,g_test_mean_r8)
    ;lea eax, g_test_sdev
    ;call print_float
    
    mov eax, offset g_test_min
    call print_stat_listing
    ret
RunTest ENDP


;; by dedndave
RunTestImpl PROC    lpfnFunction:LPVOID,lpszString:LPSTR

        mov     eax,lpfnFunction
        mov     [gu_testfunction],eax
        push    ebp
        call    run_driver
        pop     ebp
        push    eax     ;save return value

        pop     eax     ;EAX = return value
        pop     edx
        ret

RunTestImpl ENDP

RunBaseline PROC lpfnFunction:LPVOID,lpszString:LPSTR,innerLoopSize:DWORD
  ;retry:
    mov [g_isbaseline], 1
    mov eax, [innerLoopSize]
    mov [g_baseline_inner_loop_count], eax
    invoke RunTestImpl,lpfnFunction,lpszString
  ;  cmp [g_baseline_range], 8
  ;  jg retry

    ret
RunBaseline ENDP

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;; YOUR CODE TO TEST GOES HERE!
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
include ConstDiv.inc

.data

.code
align 16

;; The purpose of the baseline is to filter out overlapping work.
;; If the algorithms under test need to setup stuff prior to running the actual
;; algorithm, then this function should also do the same!
;; We generally include a loop to help amortize the overhead of
;; actually calling the function.
testit_baseline proc
    push ebx
    mov ecx, gu_inner_loop_count-1
  l:
    mov eax, 9999999
    dec ecx
    jns l
    pop ebx
    ret
testit_baseline endp


align 16
testit_cdiv proc
    push ebx
    mov ecx, gu_inner_loop_count-1
  l:
    mov eax, 9999999
    cdiv 100000
    dec ecx
    jns l
    pop ebx
    ret
testit_cdiv endp

align 16
testit_div proc
    push ebx
    mov ecx, gu_inner_loop_count-1
  l:
    mov eax, 9999999
    xor edx,edx
    mov ebx, 100000
    div ebx
    dec ecx
    jns l
    pop ebx
    ret
testit_div endp


;; a longer baseline, this time with an inner loop 128 times long.
align 16
testit_long_baseline proc
    push ebx
    mov ecx, gu_long_inner_loop_count-1
  l:
    mov eax, 9999999
    dec ecx
    jns l
    pop ebx
    ret
testit_long_baseline endp

align 16
long_testit_cdiv proc
    push ebx
    mov ecx, gu_long_inner_loop_count-1
  l:
    mov eax, 9999999
    cdiv 100000
    dec ecx
    jns l
    pop ebx
    ret
long_testit_cdiv endp

align 16
long_testit_div proc
    push ebx
    mov ecx, gu_long_inner_loop_count-1
  l:
    mov eax, 9999999
    xor edx,edx
    mov ebx, 100000
    div ebx
    dec ecx
    jns l
    pop ebx
    ret
long_testit_div endp

;; A shorter baseline
align 16
short_testit_baseline proc
    push ebx
    mov ecx, gu_short_inner_loop_count-1
  l:
    mov eax, 9999999
    dec ecx
    jns l
    pop ebx
    ret
short_testit_baseline endp

align 16
short_testit_cdiv proc
    push ebx
    mov ecx, gu_short_inner_loop_count-1
  l:
    mov eax, 9999999
    cdiv 100000
    dec ecx
    jns l
    pop ebx
    ret
short_testit_cdiv endp

align 16
short_testit_div proc
    push ebx
    mov ecx, gu_short_inner_loop_count-1
  l:
    mov eax, 9999999
    xor edx,edx
    mov ebx, 100000
    div ebx
    dec ecx
    jns l
    pop ebx
    ret
short_testit_div endp


align 16
runner proc

    
    invoke RunBaseline,testit_long_baseline,chr$("testit_long_baseline"),
           gu_long_inner_loop_count
    invoke RunTest,long_testit_div, chr$("long_testit_div")
    invoke RunTest,long_testit_cdiv, chr$("long_testit_cdiv")

    invoke RunBaseline,testit_baseline,chr$("testit_baseline"),
           gu_inner_loop_count
    invoke RunTest,testit_div, chr$("testit_div")
    invoke RunTest,testit_cdiv, chr$("testit_cdiv")

    invoke RunBaseline,short_testit_baseline,chr$("short_testit_baseline"),
           gu_short_inner_loop_count
    invoke RunTest,short_testit_div, chr$("short_testit_div")
    invoke RunTest,short_testit_cdiv, chr$("short_testit_cdiv")

    ret
runner endp

; 

end start
